/**
 * Hunter Adams (vha3@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Serial console on GPIO 0 and 1 for debugging
 * 
 * Joystick: 
 * HiLetgo Game Joystick Sensor Game Controller -- 
 * https://www.amazon.com/dp/B00P7QBGD2?psc=1&ref=ppx_yo2ov_dt_b_product_details
 * joystick 
 * pins:
 * GND ---> PICO GND
 * +5 ---> PICO +3
 * Rx ---> ADC channel 0 GPIO 26
 * Ry ---> ADC channel 0 GPIO 27
 * SW ---> GPIO 4 with pullup turned on
 * 
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
// #include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/adc.h"
// // Our assembled programs:
// // Each gets the name <pio_filename.pio.h>
// #include "hsync.pio.h"
// #include "vsync.pio.h"
// #include "rgb.pio.h"

// ==========================================
// === hardware and protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

short adc_x, adc_y ;
short draw_x, draw_y, last_draw_x, last_draw_y, last_click_x, last_click_y, x_latch ;
short click, double_click, button, last_button, click_latch, double_click_latch ;
short first_click = 1;
int click_time, last_click_time ;
int clicked_item ;

// ===========================================
// ADC setup 
// ===========================================
// setup ADC
void ADC_setup(void){
    adc_init();
    // Make sure GPIO is high-impedance, no pullups etc
    // x
    adc_gpio_init(26);
    // y
    adc_gpio_init(27);
}

// ===========================================
// Cursor  
// ===========================================
short cursor_pixel_x[5], cursor_pixel_y[5] ;

// replace cursor color with saved image colors
void erase_cursor(short x, short y){
    // first the x direction, then y
    for(int i=0; i<5; i++){
        drawPixel(x-2+i, y,  cursor_pixel_x[i]) ;
        drawPixel(x, y-2+i,  cursor_pixel_y[i]) ;
    }
}

// replace image with cursor color and save image pixels
void draw_cursor(short x, short y){
    for(int i=0; i<5; i++){
        cursor_pixel_x[i] = readPixel(x-2+i, y) ;
        cursor_pixel_y[i] = readPixel(x, y-2+i) ;
        drawPixel(x-2+i, y, WHITE) ;
        drawPixel(x, y-2+i, WHITE) ;
    }
}

// ==================================================
// === set up a menu scheme for use with VGA
// ==================================================
// ALL items store and update using float values
// int values are copied from the float
struct menu_item {
    // displayed string
    char item_name[15];
    // displayed color
    int item_color ;
    //1=float 0=int ;  
    int item_data_type ;
    // 1=log 0=linear
    int item_inc_type ;
    // increment delta if linear, ratio if log
    float item_increment ;
    // the current value, min, max 
    // al updates are performed on float, then copied to int
    int item_int_value ;
    float item_float_value ;
    float item_float_min ;
    float item_float_max ;
};

// ======
// build the menu array of items
struct menu_item menu[16] ;
//
void draw_item(short menu_x, short menu_y, int item_index, int menu_num_items){
    if(item_index<menu_num_items){
        // display the name
        setTextColor(menu[item_index].item_color) ;
        setCursor(menu_x, menu_y+item_index*25) ;
        setTextSize(1) ;
        writeString(menu[item_index].item_name) ;
        // display the value
        setTextColor(WHITE) ;
        fillRect(menu_x, menu_y+item_index*25+10, 90, 11, BLACK) ;
        setCursor(menu_x, menu_y+item_index*25+11) ;
        char value_str[15] ;
        // copy float to int value
        menu[item_index].item_int_value = (int) menu[item_index].item_float_value ;
        // get the data type in bit 0
        if (menu[item_index].item_data_type == 0) 
            sprintf(value_str, "%d", menu[item_index].item_int_value) ;
        if (menu[item_index].item_data_type == 1) 
            sprintf(value_str, "%1.2e", menu[item_index].item_float_value);
        writeString(value_str) ;
    }
}

// =====
void draw_menu(short menu_x, short menu_y, int menu_num_items){
    for(int i=0; i<menu_num_items; i++){
        draw_item( menu_x, menu_y, i,  menu_num_items) ;
    }
}

// =====
// direction is 1 for increase,  -1 for decrease
void change_value(int index, int direction){
    if(direction>0){
        if(menu[index].item_inc_type==0) menu[index].item_float_value += menu[index].item_increment ;
        if(menu[index].item_inc_type==1) menu[index].item_float_value *= menu[index].item_increment ;
    }
    if(direction<0){
        if(menu[index].item_inc_type==0) menu[index].item_float_value -= menu[index].item_increment ;
        if(menu[index].item_inc_type==1) menu[index].item_float_value /= menu[index].item_increment ;
    }
    // check min/max
    if(menu[index].item_float_value > menu[index].item_float_max)
        menu[index].item_float_value = menu[index].item_float_max ;
    if(menu[index].item_float_value < menu[index].item_float_min)
        menu[index].item_float_value = menu[index].item_float_min ;
    // update integer to match
    menu[index].item_int_value = (int) menu[index].item_float_value ;
}

// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // update count -- for tuning cursor speed
    static int update_count ;
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // draw a big, clunky banner
    #define banner_bottom 50
    // Draw some filled rectangles at top of screen
    fillRect(64, 0, 176, banner_bottom, BLUE); // blue box
    fillRect(250, 0, 176, banner_bottom, YELLOW); // red box
    fillRect(435, 0, 176, banner_bottom, GREEN); // green box

    // Write some intro banner text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("Graphics joystick demo") ;
    setCursor(65, 20) ;
    writeString("Hunter Adams & Bruce Land") ;
    setCursor(65, 30) ;
    writeString("vha3@cornell.edu") ;
    setCursor(65, 40) ;
    writeString("brl4@cornell.edu") ;
    
    setTextColor(BLACK) ;
    setCursor(290, 6) ;
    setTextSize(2) ;
    writeString("ECE 4760") ;

    setCursor(445, 10) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.1.1") ;
    setCursor(445, 20) ;
    writeString("VGA - 640x480 - 16 colors ") ;
    setTextColor(WHITE) ;

    // draw a menu outline
    #define menu_side 100
    #define item_height 25
    // left edge
    drawLine(0, banner_bottom+22, 0, 472, RED) ;
    // right edge
    drawLine(menu_side, banner_bottom+22, menu_side, 472, RED) ;
    // item dividers
    for (short i=banner_bottom+22; i<480; i+=item_height) {
        drawLine(0, i, menu_side, i, RED) ;
    }

    // test menu setup
    // bulky, but unavoidable
    menu[0].item_color = WHITE ;
    menu[1].item_color = GREEN ;
    menu[2].item_color = ORANGE ;
    menu[3].item_color = WHITE ;
    //
    sprintf(menu[0].item_name, "Draw Color ") ;
    sprintf(menu[1].item_name, "Int, log ") ;
    sprintf(menu[2].item_name, "Float, lin ") ;
    sprintf(menu[3].item_name, "Float, log ") ;
    //
    menu[0].item_data_type = 0 ;// int
    menu[1].item_data_type = 0 ;// int
    menu[2].item_data_type = 1 ;// float
    menu[3].item_data_type = 1 ;// float
    //
    menu[0].item_inc_type = 0 ;// lin update
    menu[1].item_inc_type = 1 ;// log update
    menu[2].item_inc_type = 0 ;// lin
    menu[3].item_inc_type = 1 ;// log
    //
    menu[0].item_increment = 1 ; 
    menu[1].item_increment = 2 ;
    menu[2].item_increment = .1 ;
    menu[3].item_increment = 1.1 ; 
    // 
    menu[0].item_float_value = 10 ;// 
    menu[1].item_float_value = 3000 ;// 
    menu[2].item_float_value = 1.5 ;// 
    menu[3].item_float_value = 1000 ;// 

    menu[0].item_float_min = 0 ;
    menu[0].item_float_max = 15 ;
    menu[1].item_float_min = 1 ;
    menu[1].item_float_max = 10000 ;
    menu[2].item_float_min = 0 ;
    menu[2].item_float_max = 10 ;
    menu[3].item_float_min = 100 ;
    menu[3].item_float_max = 2000 ;
   // 
   draw_menu(3, banner_bottom+26, 4);
    
    // init the ADC
    ADC_setup();

    // set up gpio 4 for joystick button
    #define joystick_button 4
     gpio_init(joystick_button) ;	
     gpio_set_dir(joystick_button, GPIO_IN) ;
     // pullup ON, pulldown OFF
     gpio_set_pulls (joystick_button, true, false) ;

     // init drawing
     draw_x = 110 ;
     draw_y = 100 ;
     last_draw_x = draw_x ;
     last_draw_y = draw_y ;

    while(true) {
        // count passes thru loop
        update_count++ ;
        // read x, y potentiometers
        adc_select_input(0);
        adc_x = adc_read();
        adc_select_input(1);
        adc_y = adc_read();
        // read the joystick pushbutton
        last_button = button ;
        button = !gpio_get(joystick_button);

        // derive a 'click' for one pass through the following code
        if(click==false && button==true && last_button==false) {
            click = true;
            last_click_time = click_time ;
            click_time = PT_GET_TIME_usec();
        }
        else click = false;

        // get double click
        #define max_double_time 500000
        if(click==true && (click_time-last_click_time)<max_double_time) double_click = true;
        else double_click = false ;
        // testing
        //printf("%d %d %d %d\n\r", adc_x, adc_y, click, double_click);

        // save the old position
        last_draw_x = draw_x ;
        last_draw_y = draw_y ;
        // init the drawing position once
        if (first_click==true){
            last_click_x = draw_x ;
            last_click_y = draw_y ;
        }
        
        // get the new posiion
        // joystick dead zone and about 4 step sizes +/-
        draw_x += (adc_x-2048)/512 ;
        // bounds check
        if (draw_x > 639) draw_x=639 ;
        // check if a menu item has been clicked and keep cursor in the item
        if  ((draw_x >= menu_side) && click_latch) draw_x=menu_side-1 ;
        if (draw_x < 0) draw_x=0 ;
        // y
        draw_y += (adc_y-2048)/512 ; 
        if (draw_y > 639) draw_y=639 ;
        if (draw_y < 0) draw_y=0 ;
        
        // erase cursor BEFORE the draw stuff
        erase_cursor(last_draw_x, last_draw_y);

        // draw if below banner and right of menu
        if (click==true && draw_y>banner_bottom && draw_x>menu_side){
            // this code only draws lines
            drawLine(last_click_x, last_click_y, draw_x, draw_y, menu[0].item_int_value);
            last_click_x = draw_x ;
            last_click_y = draw_y ;
            first_click=false;
        }

        // detect click in menu region and get item number
        if (click==true && draw_y>banner_bottom && draw_x<menu_side){
        //if (click==true && ((draw_y>banner_bottom && draw_x<menu_side) | x_latch==draw_y)){
            // get item index and set update flag
            clicked_item = (draw_y - banner_bottom -22)/25 ;
            click_latch = (click_latch)? false : true;
            // save y value to make sure we are insame item later
            x_latch = draw_y ;
        }
        
        // change the clicked item
        //if(click_latch==true  && draw_y>banner_bottom && draw_x<menu_side){
        if(click_latch==true  && draw_y==x_latch){
            // slow down the inc/dec by 4x
            if((update_count & 0x03) == 0) change_value(clicked_item, (adc_x-2048)/512);
           // update the menu text
            draw_item(3, banner_bottom+26, clicked_item, 4);
            // item change indicator
            fillCircle(95, banner_bottom+26+clicked_item*25, 2, RED) ;
        }
        else { 
            fillCircle(95, banner_bottom+26+clicked_item*25, 2, BLACK) ;
            click_latch = false ;
        }
        
        // redraw cursor
        draw_cursor(draw_x, draw_y) ;
        
        // slow down update during menu operations
        //int delay_usec = (click_latch==true)?60000: 60000 ;
        // A brief nap
        PT_YIELD_usec(60000) ;
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
// just a heartbeat to make sure we did not crash
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 1
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      
      //
      while(1) {
        // print prompt
        //sprintf(pt_serial_out_buffer, "just testing:");
        // spawn a thread to do the non-blocking write
        //serial_write ;

        // spawn a thread to do the non-blocking serial read
         //serial_read ;
         //printf("returned");
         PT_YIELD_usec(100000) ;
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main